/* This file is placed in the public domain by its
   author Tibor 'Igor2' Palinkas in 2013. Feel free to
   copy and extend it in your application */

/* Example of a simple event parser based configuration loader. The lihata
   document shall contain a single hash root with the key=value pairs.
   Each value can be set only once (as lihata hash mandates); unknown
   names will cause error; settings may have default value; when marked
   as required, a missing key causes error. */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "liblihata/parser.h"

/* === config loader === */
typedef enum conf_type_s {
	INT,
	DOUBLE,
	STRING
} conf_type_t;

typedef struct config_s {
	const char *name;
	conf_type_t type;
	void *value;
	int required;
	int set;
} config_t;

#define error(text) \
	do { \
		fprintf(stderr, "'%s' (at %d:%d)\n", text, ctx->line+1, ctx->col+1); \
		exit(1); \
	} while(0)

typedef struct {
	int level;
	config_t *cfg;
} cfg_ctx_t;

void event(lht_parse_t *ctx, lht_event_t ev, lht_node_type_t nt, const char *name, const char *value)
{
	cfg_ctx_t *conf = ctx->user_data;
	config_t *c;

	switch(ev) {
		case LHT_OPEN:
			if (conf->level > 0)
				error("expected a flat has of settings (1)");
			if (nt != LHT_HASH)
				error("root node must be a hash");
			conf->level++;
			break;
		case LHT_TEXTDATA:
			if (conf->level != 1)
				error("expected a flat has of settings (2)");
			if (*name == '\0')
				error("anonymous setting is not allowed");
			/* try setting the value - slow linear list for simplicity of the example */
			for(c = conf->cfg; c->name != NULL; c++) {
				if (strcmp(c->name, name) == 0) {
					char *end;
					if (c->set)
						error("Multiple instances of a setting");
					switch(c->type) {
						case INT:
							*(int *)c->value = strtol(value, &end, 0);
							if (*end != '\0')
								error("Invalid integer");
							break;
						case DOUBLE:
							*(double *)c->value = strtod(value, &end);
							if (*end != '\0')
								error("Invalid double");
							break;
						case STRING:
							*(char **)c->value = strdup(value);
							break;
					}
					c->set++;
				}
			}
			break;
		case LHT_ERROR:
			fprintf(stderr, "*PARSE ERROR* '%s' (at %d:%d)\n", name, ctx->line+1, ctx->col+1);
			exit(1);
			break;
		default:
			/* ignore anything else */
			;
	}
}

void cfg_load(config_t *cfg, const char *fn)
{
	FILE *f;
	lht_parse_t p;
	config_t *c;
	cfg_ctx_t conf;
	int err;

	f = fopen(fn, "r");
	if (f == NULL) {
		fprintf(stderr, "Can't open config file %s\n", fn);
		exit(1);
	}

	conf.cfg = cfg;
	conf.level = 0;
	lht_parser_init(&p);
	p.event = event;
	p.user_data = &conf;

	while(!(feof(f))) {
		int c = fgetc(f);
		if (lht_parser_char(&p, c))
			break; /* error or stop, do not read anymore (would get PE_STOP without any processing all the time) */
	}
	lht_parser_uninit(&p);
	fclose(f);

	/* error events caused exit(1) already - if we got here, we have finished loading a valid lihata document */
	for(err = 0, c = cfg; c->name != NULL; c++)
		if (c->required && !c->set) {
			fprintf(stderr, "Configuration error: %s is required but not set by %s\n", c->name, fn);
			err++;
		}
	if (err)
		exit(1);
}

/* ===  test app === */
int main(void)
{
	int     inum = 4;
	double  dnum = 42.42;
	char    *str = "default string";
	int     reqd;
	config_t cfg[] = {
		/* name     type     ptr     req  set */
		{"inum",    INT,     &inum,   0,   0},
		{"dnum",    DOUBLE,  &dnum,   0,   0},
		{"str",     STRING,  &str,    0,   0},
		{"reqd",    INT,     &reqd,   1,   0},
		{NULL,      0,       NULL,    0,   0}
	};

	cfg_load(cfg, "cfg_simple.lht");
	printf("%d %f '%s' %d\n", inum, dnum, str, reqd);
	return 0;
}
