/*
    liblihata - list/hash/table format, tree tester tool
    Copyright (C) 2013  Tibor 'Igor2' Palinkas

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

    Project URLs: http://repo.hu/projects/lihata
                  svn://repo.hu/lihata


   This file contains a minimalistic interpreter for varrying out tests
   on the lht_tree* functionality.
*/

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <assert.h>
#include "dom.h"
#include "tree.h"

lht_doc_t *doc[32];
lht_node_t *node[32];


void load_file(lht_doc_t **doc, const char *fn)
{
	FILE *f;

	f = fopen(fn, "r");
	if (f == NULL) {
		fprintf(stderr, "can not open file '%s' for read\n", fn);
		exit(1);
	}
	*doc = lht_dom_init();
	while(!(feof(f))) {
		lht_err_t err;
		int c = fgetc(f);
		err = lht_dom_parser_char(*doc, c);
		if (err != LHTE_SUCCESS) {
			if (err != LHTE_STOP) {
				fprintf(stderr, "* error in '%s': %s *\n", fn, lht_err_str(err));
				exit(1);
			}
			break; /* error or stop, do not read anymore (would get LHTE_STOP without any processing all the time) */
		}
	}
	fclose(f);
}

static void shift_ws(char **args)
{
	while(isspace(**args)) (*args)++;
}

static char *shift_str(char **args)
{
	char *s, *start;
	int len;

	start = *args;
	while(!isspace(**args)) (*args)++;
	len = *args - start;
	s = malloc(len+1);
	memcpy(s, start, len);
	s[len] = '\0';
	return s;
}

static int shift_int(char **args)
{
	char *end;
	int n;

	n = strtol(*args, &end, 10);
	if ((*end != '\0') && (*end != ' ') && (*end != '\t')) {
		fprintf(stderr, "expected an integer, got something else ('%s')\n", *args);
		exit(1);
	}

	*args = end;
	return n;
}

static int shift_id(char **args, char *type, char idchar, int arrsize)
{
	char *end;
	int id;

	shift_ws(args);
	if (**args == '\0') {
		fprintf(stderr, "expected a %s ID, got nothing\n", type);
		exit(1);
	}
	if (**args != idchar) {
		fprintf(stderr, "expected a %s ID, got something else ('%s'; ID problem)\n", type, *args);
		exit(1);
	}
	(*args)++;

	id = strtol(*args, &end, 10);
	if ((*end != '\0') && (*end != ' ') && (*end != '\t')) {
		fprintf(stderr, "expected a %s ID, got something else ('%s'; integer problem)\n", type, *args);
		exit(1);
	}

	if ((id < 0) || (id >= arrsize)) {
		fprintf(stderr, "document ID is out of range ('%s')\n", *args);
		exit(1);
	}

	*args = end;
	return id;
}

static lht_doc_t **shift_doc(char **args)
{
	return doc + shift_id(args, "document", 'D', sizeof(doc) / sizeof(lht_doc_t *));
}

static lht_node_t **shift_node(char **args)
{
	return node + shift_id(args, "node", 'N', sizeof(node) / sizeof(lht_node_t *));
}

void cmd_load_doc(char *args)
{
	lht_doc_t **doc;

	doc = shift_doc(&args);
	shift_ws(&args);
	load_file(doc, args);
}

void cmd_print_doc(char *args)
{
	lht_doc_t **doc;

	doc = shift_doc(&args);
	if ((*doc)->root == NULL)
			printf("  * empty document *\n");
		else
			lht_dom_ptree((*doc)->root, stdout, "  ");
}

void cmd_print_node(char *args)
{
	lht_node_t **node;

	node = shift_node(&args);
	if ((*node) == NULL)
			printf("  * node is NULL *\n");
		else
			lht_dom_pnode(*node, stdout, "  ");
}

void cmd_new_text(char *args)
{
	lht_node_t **node;
	char *s;

	node = shift_node(&args);
	shift_ws(&args);
	s = shift_str(&args);
	*node = lht_dom_node_alloc(LHT_TEXT, "");
	(*node)->data.text.value = s;
}

void cmd_print_tree(char *args)
{
	lht_node_t **node;

	node = shift_node(&args);
	if ((*node) == NULL)
			printf("  * node is NULL *\n");
		else
			lht_dom_ptree(*node, stdout, "  ");
}

void cmd_free_doc(char *args)
{
	lht_doc_t **doc;

	doc = shift_doc(&args);
	lht_dom_uninit(*doc);
}

void cmd_root(char *args)
{
	lht_node_t **node;
	lht_doc_t **doc;

	node = shift_node(&args);
	doc = shift_doc(&args);
	*node = (*doc)->root;
}

void cmd_list_next(char *args)
{
	lht_node_t **node;

	node = shift_node(&args);
	if ((*node)->parent == NULL) {
		fprintf(stderr, "node is not in a list (root node)\n");
		exit(1);
	}
	if ((*node)->parent->type != LHT_LIST) {
		fprintf(stderr, "node is not in a list\n");
		exit(1);
	}
	(*node) = (*node)->next;
}

void cmd_replace(char *args)
{
	lht_node_t **dst, **src;
	shift_ws(&args);
	dst = shift_node(&args);
	if ((*node)->parent == NULL) {
		fprintf(stderr, "target node is not in a documnt (detached, floating)\n");
		exit(1);
	}

	src = shift_node(&args);

	lht_tree_replace(*dst, *src);
	lht_tree_del(*dst);
}

void cmd_list_child(char *args)
{
	lht_node_t **node;

	node = shift_node(&args);
	if ((*node)->type != LHT_LIST) {
		fprintf(stderr, "node is not a list\n");
		exit(1);
	}
	(*node) = (*node)->data.list.first;
}

void cmd_list_nth(char *args)
{
	lht_node_t **node;
	int n;

	node = shift_node(&args);
	shift_ws(&args);
	n = shift_int(&args);
	*node = lht_tree_list_nth(*node, n);
}

void cmd_list_del_nth(char *args)
{
	lht_node_t **node;
	int n;

	node = shift_node(&args);
	shift_ws(&args);
	n = shift_int(&args);
	printf("list_del_nth: %d\n", lht_tree_list_del_nth(*node, n));
}

void cmd_list_del_child(char *args)
{
	lht_node_t **list, **child;

	list = shift_node(&args);
	child = shift_node(&args);
	printf("list_del_child: %d\n", lht_tree_list_del_child(*list, *child));
	*child = NULL;
}

void cmd_list_nthname(char *args)
{
	lht_node_t **node;
	int n;

	node = shift_node(&args);
	shift_ws(&args);
	n = shift_int(&args);
	shift_ws(&args);
	*node = lht_tree_list_nthname(*node, n, args);
}

void cmd_table_cell(char *args)
{
	lht_node_t **node;
	int line, col;

	node = shift_node(&args);
	shift_ws(&args);
	line = shift_int(&args);
	shift_ws(&args);
	col = shift_int(&args);
	*node = lht_dom_table_cell(*node, line, col);
}

void cmd_tree_merge(char *args)
{
	lht_node_t **dst, **src;
	lht_err_t err;

	dst = shift_node(&args);
	src = shift_node(&args);
	err = lht_tree_merge(*dst, *src);
	if (err != LHTE_SUCCESS)
		printf("merge error: %s\n", lht_err_str(err));
	else
		*src = NULL;
}

#define table_op_int(name) \
void cmd_table_ ## name(char *args) \
{ \
	lht_node_t **node; \
	int i; \
	node = shift_node(&args); \
	shift_ws(&args); \
	i = shift_int(&args); \
	fprintf(stdout, "table_" #name ": %d\n", lht_tree_table_ ## name(*node, i)); \
}

table_op_int(ins_col);
table_op_int(ins_row);
table_op_int(del_col);
table_op_int(del_row);

void cmd_path_(char *args, int follow_sy)
{
	lht_node_t **dst;
	lht_err_t err;

	dst = shift_node(&args);
	assert(dst != NULL);
	shift_ws(&args);

	*dst = lht_tree_path_((*dst)->doc, *dst, args, follow_sy, 0, &err);
	if (err != LHTE_SUCCESS)
		printf("Path error: %s\n", lht_err_str(err));
}

void cmd_path(char *args)
{
	cmd_path_(args, 1);
}

void cmd_path_nofollow(char *args)
{
	cmd_path_(args, 0);
}

void cmd_equ_node(char *args)
{
	lht_node_t **dst, **src;

	dst = shift_node(&args);
	src = shift_node(&args);

	*dst = *src;
}

void cmd_equ_doc(char *args)
{
	lht_doc_t **dst, **src;

	dst = shift_doc(&args);
	src = shift_doc(&args);

	*dst = *src;
}

void cmd_echo(char *args)
{
	printf("%s\n", args);
}

typedef struct {
	const char *name;
	void (*cmd)(char *args);
} cmd_t;

cmd_t cmds[] = {
	{"load_doc", cmd_load_doc},          /* Dtarget filename */
	{"free_doc", cmd_free_doc},          /* Dtarget */
	{"print_doc", cmd_print_doc},        /* Dsource */
	{"print_node", cmd_print_node},      /* Nsource */
	{"print_tree", cmd_print_tree},      /* Nsource */
	{"replace", cmd_replace},            /* Ndest Nsource */
	{"new_text", cmd_new_text},          /* Ntarget text */
	{"list_next", cmd_list_next},        /* Ntarget */
	{"list_child", cmd_list_child},      /* Ntarget */
	{"list_nth", cmd_list_nth},          /* Ntarget int */
	{"list_nthname", cmd_list_nthname},  /* Ntarget int name */
	{"list_del_nth", cmd_list_del_nth},  /* Nlist int */
	{"list_del_child", cmd_list_del_child},/* Nlist Nchild */
	{"table_cell", cmd_table_cell},       /* Ntable line col */
	{"table_ins_row", cmd_table_ins_row}, /* Ntable int */
	{"table_ins_col", cmd_table_ins_col}, /* Ntable int */
	{"table_del_row", cmd_table_del_row}, /* Ntable int */
	{"table_del_col", cmd_table_del_col}, /* Ntable int */
	{"tree_merge", cmd_tree_merge},       /* Ndst Nsrc */
	{"path", cmd_path},                   /* Ndest str */
	{"path_nofollow", cmd_path_nofollow}, /* Ndest str */
	{"echo", cmd_echo},                  /* string */
	{"root", cmd_root},                  /* Ntarget Dsource */
	{"equ_node", cmd_equ_node},          /* Ntarget Nsource */
	{"equ_doc",  cmd_equ_node},          /* Dtarget Dsource */
	{NULL, NULL}
};

void command(char *line)
{
	char *args;
	cmd_t *c;

	args = strpbrk(line, " \t");
	if (args != NULL) {
		*args = '\0';
		args++;
	}
	shift_ws(&args);
	for(c = cmds; c->name != NULL; c++) {
		if (strcmp(c->name, line) == 0) {
			c->cmd(args);
			return;
		}
	}
	fprintf(stderr, "Invalid command %s\n", line);
	exit(1);
}

int main()
{
	memset(doc, 0, sizeof(doc));
	memset(node, 0, sizeof(node));

	while(!(feof(stdin))) {
		char line[1024], *l, *end;
		*line = '\0';
		fgets(line, sizeof(line)-1, stdin);
		l = line;
		while(isspace(*l)) l++;
		switch(*l) {
			case '\0':
			case '#':
			case '\n':
			case '\r':
				break;
			default:
				/* strip trailing whitespace */
				end = l + strlen(l) - 1;
				while((*end == '\r') || (*end == '\n')) {
					*end = '\0';
					if (end == l)
						break;
					end--;
				}
				command(l);
		}
	}

	return 0;
}
